#version 150

// DEFAULT_FRAGMENT_SHADER

#define MAX_LIGHTS 10

in vec4 vert_position_model;
in vec3 vert_normal_model;
smooth in vec2 texture_coordinate;
in mat3 tangent_matrix;
in mat3 inv_tangent_matrix;

out vec4 outputColor;

layout(std140) uniform TransformationMatrices
{
    mat4 camera_matrix;
    mat4 persp_matrix;
};

struct PointLight {
    vec4 position;
    vec4 intensity;
    vec4 specular_color;
    float constant_attenuation;
    float linear_attenuation;
    float quadratic_attenuation;
};

layout(std140) uniform PointLights
{
    PointLight point_lights[MAX_LIGHTS];
};

uniform uint num_point_lights;

struct spot {
	vec4 position;
	vec4 intensity;
    vec4 direction;
    vec4 specular_color;
    float exp;
    float constant_attenuation;
    float linear_attenuation;
    float quadratic_attenuation;
    float cos_opening;
};

layout(std140) uniform SpotLights
{
	spot spot_lights[MAX_LIGHTS];
};

uniform uint num_spot_lights;

uniform sampler2D diff_texture;
uniform sampler2D normal_map;
uniform sampler2D height_map;
uniform sampler2D specular_map;
uniform mat4 model_matrix;
uniform mat4 inv_model_matrix;
uniform vec4 eye_coordinates;
uniform float tile_factor;
uniform float depth_scale_factor;
uniform float shininess;

float ray_intersect_search(in sampler2D height_map, in vec2 dp, in vec2 ds)
{
	const int linear_search_steps = 30;
	const int binary_search_steps = 25;
	float depth_step = 1.0f / linear_search_steps;
	float dsize = depth_step; 
	float depth = 0.0f; 
	float best_depth = 1.0f;

	for ( int i = 0; i < linear_search_steps - 1; i++){
		depth += dsize;
		float t = texture(height_map, dp + ds * depth).x;
		if (best_depth >= 0.996f){
			if (depth >= t){
				best_depth = depth;
			}
		} 
	}
	depth = best_depth;
	for ( int i = 0; i < binary_search_steps; i++){
		dsize *= 0.5f;
		float t = texture(height_map, dp + ds * depth).x;
		if (depth >= t){
			best_depth = depth;
			depth -= 2.0f * dsize;
		}
		depth += dsize;
	}
	return (best_depth);
}

bool is_shadowed(in sampler2D height_map, in vec2 dp, in vec2 ds)
{
	const int linear_search_steps = 30;
	float depth_step = 1.0f / linear_search_steps;
	float dsize = depth_step;
	float depth = 0.0f;
	bool isShadowed = false;
	
	for (int i = 0; i < linear_search_steps; i++){
		depth += dsize;
		float t = texture(height_map, dp + ds * depth).x;
		if (depth <= t-1.0f){
			isShadowed = true;
		}
	}
	return isShadowed;
}


void main()
{
	
	/* Initialize Shader-wide-variables */
    vec4 l_intinsity_sum = vec4(0.0f,0.0f,0.0f,1.0f);
    vec4 l_specular_sum = vec4(0.0f, 0.0f, 0.0f, 1.0f);
    vec4 diffuse_color = vec4(0.0f, 0.0f, 0.0f, 1.0f);
    
    vec4 spec_intensity = vec4(1.0f);
    
    //float shininess = 0.0f;
    vec4 fragment_position = vert_position_model;
    vec3 fragment_normal = vert_normal_model;
    
    /*Debugging relief mapping*/
    bool reliefMapping = true;

    
    /* Initialize Raytracer */
    vec3 eye_dir_model = vec3((inv_model_matrix * eye_coordinates) - fragment_position);
    vec3 eye_dir =  normalize(tangent_matrix * eye_dir_model);
    vec2 delta_s = eye_dir.xy * depth_scale_factor / eye_dir.z;
    vec2 delta_p = texture_coordinate * tile_factor;
    
    float d = ray_intersect_search(height_map, delta_p, delta_s);
	
	/* Initialize Shading Step */
	if (reliefMapping){
		vec2 intersection_point = delta_p + delta_s * d;
		diffuse_color = texture(diff_texture, intersection_point);
    	fragment_normal = normalize(inv_tangent_matrix * ((texture(normal_map, intersection_point).rgb*2.0f)-1.0f));
    	fragment_position += vec4(inv_tangent_matrix * (eye_dir * d), 1.0f);
    	spec_intensity = texture(specular_map, intersection_point);    	
    } else {
    		diffuse_color = texture(diff_texture, texture_coordinate);
    		spec_intensity = texture(specular_map, texture_coordinate);
    		
    }
    
    /* Computing Lighting from Pointslights */
    for (int i=0; i<int(num_point_lights); i++){
    	float exponent = 0.0f;
    	vec3 light_dir = vec3((inv_model_matrix*point_lights[i].position) - fragment_position);
    	float l = length(light_dir);
    	light_dir = normalize(light_dir);
    	
    	vec2 ldelta_p = delta_p + delta_s * d;
    	vec3 s = normalize(tangent_matrix * light_dir);
    	vec2 ldelta_s = (s.xy * depth_scale_factor) / s.z;
    	ldelta_p -= ldelta_s * d;
    	float point_attenuation = 0.0f;
    	float incidence = 0.0f;
    	if (is_shadowed(height_map, ldelta_p, ldelta_s)){
    
    	} else {
    	
    		point_attenuation = 1.0f / (point_lights[i].constant_attenuation + point_lights[i].linear_attenuation*l + point_lights[i].quadratic_attenuation*l*l);
    	
    		vec3 camera_pos = normalize(vec3((inv_model_matrix * eye_coordinates) - fragment_position));
    		vec3 half_angle = normalize(light_dir + camera_pos);
    		incidence = max(0.0f, dot(normalize(fragment_normal), light_dir));
    		float half_incidence = max(0.0, dot(normalize(fragment_normal), half_angle));
    	
    		if (incidence == 0.0f){
    			exponent = 0.0f;
    		} else {
    			exponent = pow(half_incidence, shininess);
    		}
    	}
    	l_intinsity_sum += point_lights[i].intensity * point_attenuation * incidence;
    	l_specular_sum += point_lights[i].specular_color * spec_intensity * point_attenuation * exponent;
    	
    }
    
    /* Computing Lighting from SpotLights */
    for (int i=0; i<int(num_spot_lights); i++){
    
	    float spotAttenuation = 0.0f;
	    float attenuation = 0.0f;
	    float exponent = 0.0f;
	    float normalProduct = 0.0f;
	   
	    vec3 spot_dir = vec3((inv_model_matrix*spot_lights[i].position) - fragment_position);
	   
	    float l = length(spot_dir);
	    spot_dir = normalize(spot_dir);
	    
	    vec2 ldelta_p = delta_p + delta_s * d;
    	vec3 s = normalize(tangent_matrix * spot_dir);
    	vec2 ldelta_s = (s.xy * depth_scale_factor) / s.z;
    	ldelta_p -= ldelta_s * d;
    	if (is_shadowed(height_map, ldelta_p, ldelta_s)){
    	} else {
	   
	    	attenuation = 1.0 / (spot_lights[i].constant_attenuation + spot_lights[i].linear_attenuation*l + spot_lights[i].quadratic_attenuation*l*l);
	   
	    	float spotEffect = dot(-spot_dir,normalize(vec3(inv_model_matrix*spot_lights[i].direction))); 
	    	if (spotEffect < spot_lights[i].cos_opening){
	        	spotAttenuation = 0.0f;
	    	}else{
	        	spotAttenuation = pow(spotEffect,spot_lights[i].exp);
	    	}
	    	attenuation *= spotAttenuation;
	   
	    	vec3 camera_pos = normalize(vec3((inv_model_matrix * eye_coordinates) - fragment_position));
    		vec3 half_angle = normalize(spot_dir + camera_pos);
	    	normalProduct = max(0.0, dot(normalize(fragment_normal), spot_dir));
	    	float half_normal_spot = max(0.0, dot(normalize(fragment_normal), half_angle));
	    	if (normalProduct == 0.0f){
    			exponent = 0.0f;
    		} else {
    			exponent = pow(half_normal_spot, shininess);
    		}
	    }
	    l_intinsity_sum += spot_lights[i].intensity * attenuation * normalProduct;
	    l_specular_sum += spot_lights[i].specular_color * spec_intensity * attenuation * exponent;
	    
    }  
    /* Apply all Intensities */
    outputColor = diffuse_color * l_intinsity_sum + l_specular_sum;
}
